#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>

#include "type.h"
#include "response.h"

/* global functions */
int get_id(WORK_BLOCK *work);
int get_block(WORK_BLOCK *work);
int report(WORK_BLOCK *work);

/* file local functions */
static int open_socket(char *host, int port);
static int pump_response(int sock, RAW_RESPONSE *raw);
static int setup_block_by_key_found(WORK_BLOCK *work, LV1_RESPONSE *lv1);
static int setup_block_by_ok(WORK_BLOCK *work, LV1_RESPONSE *lv1);
static void setup_unit(WORK_BLOCK *block, char *p);

/*----------------------------------------------------------------*/
static char  *TARGET_HOST = "www.marumo.ne.jp";
static char  *TARGET_PATH = "/c2/bf/3rd.cgi";
static short  TARGET_PORT = 80;
static char  *USER_AGENT  = "C2 Brute Force Attacker (UNIX)";
/*----------------------------------------------------------------*/

int get_id(WORK_BLOCK *work)
{
	int sock;
	int n;
	char head[1024];
	char body[1024];
	
	RAW_RESPONSE raw;
	LV1_RESPONSE lv1;
	
	if(strlen(work->proxy_host)){
		sock = open_socket(work->proxy_host, work->proxy_port);
	}else{
		sock = open_socket(TARGET_HOST, TARGET_PORT);
	}
	if(sock < 0){
		fprintf(stdout, "failed on open_socket()\n");
		return -1;
	}

	sprintf(body, "cmd=GET_ID");
	n = strlen(body);
	if(strlen(work->proxy_host)){
		sprintf(head, "POST http://%s%s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_HOST, TARGET_PATH, USER_AGENT, n);
	}else{
		sprintf(head, "POST %s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_PATH, USER_AGENT, n);
	}
	
	n = strlen(head);
	if(write(sock, head, n) != n){
		fprintf(stdout, "failed on write(sock, head)\n");
		close(sock);
		return -1;
	}
	n = strlen(body);
	if(write(sock, body, n) != n){
		fprintf(stdout, "failed on write(sock, head)\n");
		close(sock);
		return -1;
	}

	n = pump_response(sock, &raw);
	close(sock);
	
	if(n == 0){
		fprintf(stdout, "failed on pump_response\n");
		return -1;
	}

	n = setup_response_lv1(&raw, &lv1);
	if(n == 0){
		fprintf(stdout, "failed on setup_response_lv1\n");
		free(raw.data);
		return 0;
	}

	n = http_status_code(&lv1);
	if(n != 200){
		fprintf(stdout, "http_status_code: %d\n", n);
		teardown_response_lv1(&lv1);
		free(raw.data);
		return -1;
	}

	if(lv1.body_line < 1){
		fprintf(stdout, "no response body\n");
		teardown_response_lv1(&lv1);
		free(raw.data);
		return -1;
	}

	if(strlen(lv1.body[0]) != 32){
		fprintf(stdout, "body size: %d\n", strlen(lv1.body[0]));
		teardown_response_lv1(&lv1);
		free(raw.data);
		return 0;
	}

	memcpy(work->id, lv1.body[0], 32);
	work->id[32] = '\0';
	teardown_response_lv1(&lv1);
	free(raw.data);

	return 1;
}

int get_block(WORK_BLOCK *work)
{
	int sock;
	int i,j,n;
	char *p;
	
	char head[1024];
	char body[1024];
	
	RAW_RESPONSE raw;
	LV1_RESPONSE lv1;

	if(strlen(work->proxy_host)){
		sock = open_socket(work->proxy_host, work->proxy_port);
	}else{
		sock = open_socket(TARGET_HOST, TARGET_PORT);
	}
	if(sock < 0){
		fprintf(stdout, "failed on open_socket()\n");
		return -1;
	}
	
	sprintf(body, "cmd=GET_BLOCK&id=%s", work->id);
	n = strlen(body);
	if(strlen(work->proxy_host)){
		sprintf(head, "POST http://%s%s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_HOST, TARGET_PATH, USER_AGENT, n);
	}else{
		sprintf(head, "POST %s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_PATH, USER_AGENT, n);
	}
	
	n = strlen(head);
	if(write(sock, head, n) != n){
		fprintf(stdout, "failed on write(sock, head)\n");
		close(sock);
		return -1;
	}
	n = strlen(body);
	if(write(sock, body, n) != n){
		fprintf(stdout, "failed on write(sock, body)\n");
		close(sock);
		return -1;
	}

	n = pump_response(sock, &raw);
	close(sock);

	if(n == 0){
		fprintf(stdout, "failed on pump_response()\n");
		return -1;
	}

	n = setup_response_lv1(&raw, &lv1);
	if(n == 0){
		fprintf(stdout, "failed on setup_response_lv1()\n");
		free(raw.data);
		return 0;
	}

	n = http_status_code(&lv1);
	if(n != 200){
		fprintf(stdout, "http_status_code: %d\n", n);
		teardown_response_lv1(&lv1);
		free(raw.data);
		return -1;
	}

	if(lv1.body_line < 1){
		fprintf(stdout, "body_line: %d\n", lv1.body_line);
		teardown_response_lv1(&lv1);
		free(raw.data);
		return -1;
	}

	if(strncmp(lv1.body[0], "key_found", 9) == 0){
		n = setup_block_by_key_found(work, &lv1);
	}else if(strncmp(lv1.body[0], "ok", 2) == 0){
		n = setup_block_by_ok(work, &lv1);
	}else{
		fprintf(stdout, "body[0] = %s\n", lv1.body[0]);
		n = 0;
	}

	teardown_response_lv1(&lv1);
	free(raw.data);
	
	return n;
}

int report(WORK_BLOCK *work)
{
	int sock;
	int i,j,n;

	char body[4096];
	char head[1024];
	char *p;

	RAW_RESPONSE raw;
	LV1_RESPONSE lv1;

	if(strlen(work->proxy_host)){
		sock = open_socket(work->proxy_host, work->proxy_port);
	}else{
		sock = open_socket(TARGET_HOST, TARGET_PORT);
	}
	if(sock < 0){
		return -1;
	}
	
	sprintf(body, "cmd=REPORT&id=%s", work->id);
	p = body + strlen(body);
	for(i=0;i<work->count;i++){
		sprintf(p, "&block=%06x", work->unit[i].ka);
		p = p + strlen(p);
		for(j=0;j<8;j++){
			if(work->unit[i].ff & (1<<j)){
				sprintf(p, ":%016llx:%014llx", work->unit[i].c[j], work->unit[i].k[j]);
				p = p + strlen(p);
			}
		}
	}

	n = strlen(body);
	if(strlen(work->proxy_host)){
		sprintf(head, "POST http://%s%s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_HOST, TARGET_PATH, USER_AGENT, n);
	}else{
		sprintf(head, "POST %s HTTP/1.0\r\nUser-Agent: %s\r\nContent-Length: %d\r\n\r\n", TARGET_PATH, USER_AGENT, n);
	}

	n = strlen(head);
	if(write(sock, head, n) != n){
		fprintf(stdout, "failed on write(sock, head)\n");
		close(sock);
		return -1;
	}

	n = strlen(body);
	if(write(sock, body, n) != n){
		fprintf(stdout, "failed on write(sock, body)\n");
		close(sock);
		return -1;
	}

	n = pump_response(sock, &raw);
	close(sock);

	if(n == 0){
		fprintf(stdout, "failed on pump_response()\n");
		return -1;
	}

	n = setup_response_lv1(&raw, &lv1);
	if(n == 0){
		fprintf(stdout, "failed on setup_response_lv1()\n");
		free(raw.data);
		return 0;
	}

	n = http_status_code(&lv1);
	if(n != 200){
		fprintf(stdout, "http_status_code: %d\n", n);
		teardown_response_lv1(&lv1);
		free(raw.data);
		return -1;
	}

	if(lv1.body_line < 1){
		fprintf(stdout, "body_line: %d\n", lv1.body_line);
		teardown_response_lv1(&lv1);
		free(raw.data);
		return 0;
	}

	for(i=1;i<lv1.body_line;i++){
		if(strncmp(lv1.body[i], "rank=", 5) == 0){
			strcpy(work->rank, lv1.body[i]+5);
		}
	}

	if(strncmp(lv1.body[0], "ok", 2) == 0){
		teardown_response_lv1(&lv1);
		free(raw.data);
		return 1;
	}

	fprintf(stdout, "body[0] = %s\n", lv1.body[0]);
	teardown_response_lv1(&lv1);
	free(raw.data);
	
	return 0;
}

/* --------------------------------------------------------------- */

static int open_socket(char *host, int port)
{
	struct hostent *hostent;
	struct sockaddr_in addr;
	int addr_len;
	int s;

	addr.sin_family = AF_INET;
	if((hostent = gethostbyname(host)) == NULL){
		fprintf(stdout,"unknown host %s\n", host);
		return(-1);
	}
	bcopy(hostent->h_addr, &addr.sin_addr, hostent->h_length);
	addr.sin_port = htons(port);
	if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		perror("socket");
		return(-1);
	}
	if(connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0){
		perror("hostname");
		close(s);
		return(-1);
	}
	return(s);
}

static int pump_response(int sock, RAW_RESPONSE *raw)
{
	int n;
	int m;
	char *p;

	n = 4096;
	raw->size = 0;
	raw->data = (char *)malloc(n);
	if(raw->data == NULL){
		return 0;
	}
	
	while( (m = read(sock, raw->data+raw->size, n-raw->size)) > 0 ) {
		raw->size += m;
		if(raw->size == n){
			p = realloc(raw->data, n*2);
			if(p == NULL){
				break;
			}
			n *= 2;
			raw->data = p;
		}
	}

	if(raw->data[raw->size-1] != '\0'){
		if(raw->size == n){
			p = realloc(raw->data, n+1);
			if(p == NULL){
				raw->size -= 1;
			}else{
				raw->data = p;
			}
		}
		raw->data[raw->size] = '\0';
	}

	return raw->size;
}


static int setup_block_by_key_found(WORK_BLOCK *work, LV1_RESPONSE *lv1)
{
	int i;
	char *p;

	for(i=1;i<lv1->body_line;i++){
		if(strncmp(lv1->body[i], "pt=", 3) == 0){
			work->plaintext = strtoull(lv1->body[i]+3, NULL, 16);
		}else if(strncmp(lv1->body[i], "ct=", 3) == 0){
			work->unit[0].c[0] = strtoull(lv1->body[i]+3, NULL, 16);
		}else if(strncmp(lv1->body[i], "key=", 4) == 0){
			work->unit[0].k[0] = strtoull(lv1->body[i]+4, NULL, 16);
		}
	}

	work->count = 1;
	for(i=1;i<8;i++){
		work->unit[0].c[i] = work->unit[0].c[0];
		work->unit[0].k[i] = work->unit[0].k[0];
	}
	work->unit[0].ff = 0xff;
	work->unit[0].st = 2;
	work->unit[0].ka = (uint32_t)((work->unit[0].k[0]>>32)&0x00ffffff);
	work->unit[0].kb = (uint32_t)( work->unit[0].k[0]     &0xffffffff);

	return 2;
}

static int setup_block_by_ok(WORK_BLOCK *work, LV1_RESPONSE *lv1)
{
	int i;

	for(i=1;i<lv1->body_line;i++){
		if(strncmp(lv1->body[i], "pt=", 3) == 0){
			work->plaintext = strtoull(lv1->body[i]+3, NULL, 16);
		}else if(strncmp(lv1->body[i], "block=", 6) == 0){
			setup_unit(work, lv1->body[i]+6);
		}
	}

	if(work->count){
		work->time = time(NULL);
		return 1;
	}

	return 0;
}


static void setup_unit(WORK_BLOCK *work, char *p)
{
	int i,j,k;

	if(strlen(p) != 142){
		return;
	}
	
	for(i=6;i<142;i+=17){
		if(p[i] != ':'){
			return;
		}
		p[i] = '\0';
	}

	i = work->count;
	if(i >= 8){
		return;
	}

	work->unit[i].kb = 0;
	work->unit[i].ff = 0;
	work->unit[i].st = 0;
	
	work->unit[i].ka = strtoul(p, NULL, 16);
	for(j=0,k=6;j<8;j++,k+=17){
		work->unit[i].c[j] = strtoull(p+k+1, NULL, 16);
		work->unit[i].k[j] = 0;
	}
	
	work->count += 1;
}
