#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "type.h"

/* global functions */
int  load(WORK_BLOCK *wb);
void save(WORK_BLOCK *wb);

/* file local functions */
static void setup_rc_path(char *path);
static char *search_non_space(char *p);
static char *next_token(char *p);
static void parse_unit(WORK_BLOCK *wb, FILE *fp, uint32_t ka);
static int parse_code(WORK_UNIT *wu, FILE *fp, uint64_t code, int idx);
static void setup_proxy(WORK_BLOCK *wb);

/* implementations */
int load(WORK_BLOCK *wb)
{
	FILE *fp;
	int n;
	char *p;
	char buffer[1024];

	setup_rc_path(buffer);

	fp = fopen(buffer, "r");
	if(fp == NULL){
		setup_proxy(wb);
		return 0;
	}

	strcpy(wb->rank, "unknown");

	while( fgets(buffer, sizeof(buffer), fp) != NULL ){
		p = search_non_space(buffer);
		if(strncmp("id", p, 2) == 0){
			memcpy(wb->id, next_token(p+3), 32);
			wb->id[32] = 0;
		}else if(strncmp("plaintext", p, 9) == 0){
			wb->plaintext = strtoull(next_token(p+10), NULL, 16);
		}else if(strncmp("ka", p, 2) == 0){
			parse_unit(wb, fp, strtoul(next_token(p+3), NULL, 16));
		}else if(strncmp("time", p, 4) == 0){
			wb->time = strtoul(next_token(p+5), NULL, 16);
		}
	}

	fclose(fp);

	setup_proxy(wb);

	if(strlen(wb->id) == 0){
		return 0;
	}

	return 1;
}

void save(WORK_BLOCK *wb)
{
	uint32_t i,j;
	
	FILE *fp;
	
	char buffer[1024];

	setup_rc_path(buffer);

	fp = fopen(buffer, "w");
	if(fp == NULL){
		return;
	}

	fprintf(fp, "id=%s\n", wb->id);
	fprintf(fp, "time=%08x\n", wb->time);
	fprintf(fp, "plaintext=%016llx\n", wb->plaintext);
	for(i=0;i<wb->count;i++){
		fprintf(fp, "ka=%06x {\n", wb->unit[i].ka);
		fprintf(fp, "  status=%d\n", wb->unit[i].st);
		fprintf(fp, "  kb=%08x\n", wb->unit[i].kb);
		for(j=0;j<8;j++){
			fprintf(fp, "  code=%016llx {\n", wb->unit[i].c[j]);
			fprintf(fp, "    key=%016llx\n", wb->unit[i].k[j]);
			fprintf(fp, "    find=%d\n", (wb->unit[i].ff>>j)&1);
			fprintf(fp, "  }\n");
		}
		fprintf(fp, "}\n");
	}

	fclose(fp);
}

static void setup_rc_path(char *path)
{
	int n;
	char *home;

	home = getenv("HOME");
	n = strlen(home);

	if(home[n-1] != '/'){
		sprintf(path, "%s/%s", home, ".c2bfrc");
	}else{
		sprintf(path, "%s%s", home, ".c2bfrc");
	}
}

static char *search_non_space(char *p)
{
	while(isspace(*p)){
		p += 1;
	}

	return p;
}

static char *next_token(char *p)
{
	while(*p != '\0'){
		if( (!isspace(*p)) && (*p != '=')){
			break;
		}
		p += 1;
	}

	return p;
}

static void parse_unit(WORK_BLOCK *wb, FILE *fp, uint32_t ka)
{
	uint32_t idx;
	int n;
	int f;
	char *p;
	char buffer[1024];

	n = 0;
	f = 0;
	idx = wb->count;
	wb->unit[idx].ka = ka;
	while( fgets(buffer, sizeof(buffer), fp) != NULL ){
		p = search_non_space(buffer);
		if(strncmp("status", p, 6) == 0){
			f |= 1;
			wb->unit[idx].st = (uint8_t)(atoi(next_token(p+7)));
		}else if(strncmp("kb", p, 2) == 0){
			f |= 2;
			wb->unit[idx].kb = strtoul(next_token(p+3), NULL, 16);
		}else if(strncmp("code", p, 4) == 0){
			n += parse_code(wb->unit+idx, fp, strtoull(next_token(p+5), NULL, 16), n);
		}else if(p[0] == '}'){
			f |= 4;
			break;
		}
	}

	if( (f == 7) && (n == 8) ){
		wb->count += 1;
	}
}

static int parse_code(WORK_UNIT *wu, FILE *fp, uint64_t code, int idx)
{
	int f;
	char *p;
	char buffer[1024];

	f = 0;
	wu->c[idx] = code;
	while(fgets(buffer, sizeof(buffer), fp) != NULL){
		p = search_non_space(buffer);
		if(strncmp("key", p, 3) == 0){
			f |= 1;
			wu->k[idx] = strtoull(next_token(p+4), NULL, 16);
		}else if(strncmp("find", p, 4) == 0){
			f |= 2;
			if(atoi(next_token(p+5))){
				wu->ff |= (1<<idx);
			}
		}else if(p[0] == '}'){
			f |= 4;
			break;
		}
	}

	if( f == 7 ){
		return 1;
	}

	return 0;
}

static void setup_proxy(WORK_BLOCK *wb)
{
	char *p;
	char *colon;
	
	p = getenv("HTTP_PROXY");
	if(p == NULL){
		wb->proxy_host[0] = '\0';
		return;
	}

	strcpy(wb->proxy_host, p);

	colon = NULL;
	p = wb->proxy_host;
	while(*p != '\0'){
		if(*p == ':'){
			colon = p;
			break;
		}
		p += 1;
	}

	if(colon == NULL){
		wb->proxy_port = 80;
		return;
	}

	colon[0] = '\0';
	wb->proxy_port = atoi(colon+1);
}

